; hacker port firmware
; version 0.4.0 date 2004-07-30
; part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
; Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

	TITLE		"serial port procedures for hackerport firmware"

	include		"P18F442.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG
	extern		USB_Sx_CNT, USB_Sx_BUF



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		DEBUG_LOG



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; general purpose register for high priority interrupt
SER_INT_HI_GPR	res		1
; general purpose register for low priority interrupt
SER_INT_LO_GPR	res		1

; ring buffer pointers for serial port
SER_RECV_START	res		1
SER_RECV_END	res		1
SER_TRANS_START	res		1
SER_TRANS_END	res		1

; error code
SER_ERR_BITS	res		1
#define	SER_ERR_FRAME	SER_ERR_BITS,0
#define	SER_ERR_OV	SER_ERR_BITS,1
#define	SER_ERR_BUF_OV	SER_ERR_BITS,2



;===========================================================================================
ram_ser		UDATA			; normal RAM for serial port
;===========================================================================================

; backup registers for high priority interrupt
SER_INT_HI_FSR	res		2

; backup registers for low priority interrupt
SER_INT_WREG	res		1
SER_INT_STATUS	res		1
SER_INT_BSR	res		1
SER_INT_LO_FSR	res		2

; buffers for serial port
#define	SER_RECV_BUF_SIZE	24
SER_RECV_BUF	res	SER_RECV_BUF_SIZE
#define	SER_TRANS_BUF_SIZE	24
SER_TRANS_BUF	res	SER_TRANS_BUF_SIZE



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; serial port initialization
;-------------------------------------------------------------------------------------------
SER_INIT:
; set up AUSART
	movlw	b'00000110'		; 8 bit, async mode, high speed, shift reg empty
	movwf	TXSTA
	movlw	b'10000000'		; port enabled, 8 bit, recv disabled, no addr detect
	movwf	RCSTA
	movlw	21			; 115200 kbps (1,4% error)
	movwf	SPBRG
	bsf	IPR1,RCIP		; receive interrupt: high priority
	bcf	IPR1,TXIP		; transmit interrupt: low priority
	bcf	PIR1,RCIF		; clear interrupt flags
	bcf	PIR1,TXIF
; clear serial port buffers
	clrf	SER_RECV_START
	clrf	SER_RECV_END
	clrf	SER_TRANS_START
	clrf	SER_TRANS_END
; no error yet
	clrf	SER_ERR_BITS
; enable UART interrupts
	bsf	PIE1,RCIE
	bsf	PIE1,TXIE
; enable reception and transmission
	bsf	RCSTA,CREN
	bsf	TXSTA,TXEN
; done
	return

;-------------------------------------------------------------------------------------------
; AUSART receive interrupt
; uses hardware register stack
; high priority
;-------------------------------------------------------------------------------------------
SER_INT_RECV:
; check for errors
	btfsc	RCSTA,FERR
	bra	SER_INT_RECV_ERR_FRAME
	btfsc	RCSTA,OERR
	bra	SER_INT_RECV_ERR_OV
; check if space in receive buffer
	incf	SER_RECV_END,W		; get new value for end pointer
	movwf	SER_INT_HI_GPR
	movlw	SER_RECV_BUF_SIZE	; if SER_INT_HI_GPR >= SER_RECV_BUF_SIZE, ...
	cpfslt	SER_INT_HI_GPR
	clrf	SER_INT_HI_GPR		; ..., SER_INT_HI_GPR = 0
	movf	SER_RECV_START,W	; if start pointer != new end pointer, ...
	cpfseq	SER_INT_HI_GPR
	bra	SER_INT_RECV_OK		; ..., buffer is not full
; read byte (and forget it)
	movf	RCREG,W
; disable reception
	bcf	RCSTA,RCEN
; set buffer overflow error bit
	bsf	SER_ERR_BUF_OV
; clean up
	retfie	FAST
; put byte into receive buffer
SER_INT_RECV_OK:
	movff	FSR0L,SER_INT_HI_FSR+0	; backup FSR0
	movff	FSR0H,SER_INT_HI_FSR+1
	lfsr	FSR0,SER_RECV_BUF	; get address
	movf	SER_RECV_END,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	RCREG,INDF0		; get byte and put it into buffer
	movff	SER_INT_HI_FSR+0,FSR0L	; restore FSR0
	movff	SER_INT_HI_FSR+1,FSR0H
	movff	SER_INT_HI_GPR,SER_RECV_END	; set end pointer to new value
; clean up
	retfie	FAST
; set framing error bit
SER_INT_RECV_ERR_FRAME:
	bsf	SER_ERR_FRAME
	bra	SER_INT_RECV_ERR_CLR
; set overflow error bit
SER_INT_RECV_ERR_OV:
	bsf	SER_ERR_OV
; clear error condition
SER_INT_RECV_ERR_CLR:
	movf	RCREG,W			; read recieve register
	bcf	RCREG,CREN		; disable reception
; clean up
	retfie	FAST

;-------------------------------------------------------------------------------------------
; AUSART transmit interrupt
; uses interupt backup registers INT_x WREG, STATUS and BSR
; low priority
;-------------------------------------------------------------------------------------------
SER_INT_TRANS:
	movff	WREG,SER_INT_WREG	; save WREG, STATUS and BSR
	movff	STATUS,SER_INT_STATUS
	movff	BSR,SER_INT_BSR
; check if a byte is in transmit buffer
	movf	SER_TRANS_START,W	; if start pointer != end pointer, ...
	cpfseq	SER_TRANS_END
	bra	SER_INT_TRANS_OK	; ..., buffer is not empty
; disable transmission interrupt
	bcf	PIE1,TXIE
	bra	SER_INT_TRANS_END
; get byte from transmit buffer and transmit it
SER_INT_TRANS_OK:
	movff	FSR0L,SER_INT_LO_FSR+0	; backup FSR0
	movff	FSR0H,SER_INT_LO_FSR+1
	lfsr	FSR0,SER_TRANS_BUF	; get address
	movf	SER_TRANS_START,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	INDF0,TXREG		; get byte and transmit it
	movff	SER_INT_LO_FSR+0,FSR0L	; restore FSR0
	movff	SER_INT_LO_FSR+1,FSR0H
	incf	SER_TRANS_START,W	; increment start pointer
	movwf	SER_INT_LO_GPR
	movlw	SER_TRANS_BUF_SIZE	; if SER_INT_LO_GPR >= SER_TRANS_BUF_SIZE, ...
	cpfslt	SER_INT_LO_GPR
	clrf	SER_INT_LO_GPR		; ..., SER_INT_LO_GPR = 0
	movff	SER_INT_LO_GPR,SER_TRANS_START	; modify start pointer atomically
; clean up
SER_INT_TRANS_END:
	movff	SER_INT_BSR,BSR		; restore WREG, STATUS and BSR
	movff	SER_INT_STATUS,STATUS
	movff	SER_INT_WREG,WREG
	retfie

;-------------------------------------------------------------------------------------------
; set baudrate of serial port
; input:	A2_REG		bit time in 100ns
; output:	-
; changes:	WREG, STATUS, A2_REG
;-------------------------------------------------------------------------------------------
SER_SET_BAUD:
; divide by 4 to get baud rate generator value for high speed
	movlw	2			; add 2 to achieve rounded result
	addwf	A2_REG+0
	movlw	0
	addwfc	A2_REG+1
	bcf	STATUS,C		; shift right 2 bits
	rrcf	A2_REG+1
	rrcf	A2_REG+0
	bcf	STATUS,C
	rrcf	A2_REG+1
	rrcf	A2_REG+0
; if high part is zero, set baud rate in high speed mode
	movf	A2_REG+1,W
	bnz	SER_SET_BAUD_LOW
	movf	A2_REG+0,W		; if value is 0, ...
	btfsc	STATUS,Z
	incf	A2_REG+0		; ..., 	set it to 1
	bcf	TXSTA,TXEN		; disable transmission
	bcf	RCSTA,CREN		; disable reception
	bcf	PIE1,TXIE		; disable UART interrupts
	bcf	PIE1,RCIE
	bsf	TXSTA,BRGH		; high speed
	decf	A2_REG+0,W		; value - 1
	movwf	SPBRG
	clrf	SER_RECV_START		; clear serial port buffers
	clrf	SER_RECV_END
	clrf	SER_TRANS_START
	clrf	SER_TRANS_END
	bsf	PIE1,RCIE		; enable UART interrupts
	bsf	PIE1,TXIE
	bsf	RCSTA,CREN		; enable reception
	bsf	TXSTA,TXEN		; enable transmission
; done
	return
; divide by 4 to get baud rate generator value for low speed
SER_SET_BAUD_LOW:
	movlw	2			; add 2 to achieve rounded result
	addwf	A2_REG+0
	movlw	0
	addwfc	A2_REG+1
	bcf	STATUS,C		; shift right 2 bits
	rrcf	A2_REG+1
	rrcf	A2_REG+0
	bcf	STATUS,C
	rrcf	A2_REG+1
	rrcf	A2_REG+0
; set baud rate in low speed mode
	movf	A2_REG+1,W		; if high part is not 0, ...
	btfss	STATUS,Z
	clrf	A2_REG+0		; ..., set low part to 0 (will be decremented to 255)	
	bcf	TXSTA,TXEN		; disable transmission
	bcf	RCSTA,CREN		; disable reception
	bcf	PIE1,TXIE		; disable UART interrupts
	bcf	PIE1,RCIE
	bcf	TXSTA,BRGH		; low speed
	decf	A2_REG+0,W		; value - 1
	movwf	SPBRG
	clrf	SER_RECV_START		; clear serial port buffers
	clrf	SER_RECV_END
	clrf	SER_TRANS_START
	clrf	SER_TRANS_END
	bsf	PIE1,RCIE		; enable UART interrupts
	bsf	PIE1,TXIE
	bsf	RCSTA,CREN		; enable reception
	bsf	TXSTA,TXEN		; enable transmission
; done
	return

;-------------------------------------------------------------------------------------------
; queue a byte to be sent
; input:	A1_REG		the byte to be send
; output:	STATUS,C	set if buffer was full
; changes:	WREG, STATUS, FSR0, B1_REG
;-------------------------------------------------------------------------------------------
SER_SEND:
; check if space in transmit buffer
	incf	SER_TRANS_END,W		; get new value for end pointer
	movwf	B1_REG
	movlw	SER_TRANS_BUF_SIZE	; if B1_REG >= SER_TRANS_BUF_SIZE, ...
	cpfslt	B1_REG
	clrf	B1_REG			; ..., B1_REG = 0
	movf	SER_TRANS_START,W	; if start pointer != new end pointer, ...
	cpfseq	B1_REG
	bra	SER_SEND_OK		; ..., buffer is not full
; enable transmission interrupt
	bsf	PIE1,TXIE
; return error: buffer full
	bsf	STATUS,C
	return
; put byte into transmit buffer
SER_SEND_OK:
	lfsr	FSR0,SER_TRANS_BUF	; get address
	movf	SER_TRANS_END,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	A1_REG,INDF0		; write byte
	movff	B1_REG,SER_TRANS_END	; set end pointer to new value
; enable transmission interrupt
	bsf	PIE1,TXIE
; return success
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; serial port task
; - called periodically
; changes:	WREG, STATUS, FSR0, A1_REG
;-------------------------------------------------------------------------------------------
SER_TASK:
; check for error
	movf	SER_ERR_BITS,W
	bz	SER_TASK_NO_ERR
; check if enough space in message send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	return				; ..., do nothing
; disable AUSART receive interrupt
	bcf	PIE1,RCIE
; send error message
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x23			; put data into send buffer
	movwf	POSTINC0
	movff	SER_ERR_BITS,POSTINC0	; error code
; clear error bits
	clrf	SER_ERR_BITS
; clear receive buffer
	clrf	SER_RECV_START
	clrf	SER_RECV_END
; enable AUSART receive interrupt
	bsf	PIE1,RCIE
; enable reception
	bsf	RCSTA,CREN
; done
	return
SER_TASK_NO_ERR:
; check if a byte is in receive buffer
	movf	SER_RECV_START,W	; if start pointer == end pointer, ...
	xorwf	SER_RECV_END,W
	btfsc	STATUS,Z
	return				; ..., buffer is empty, do nothing
; check if enough space in message send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	return				; ..., do nothing
; get byte from receive buffer
	lfsr	FSR0,SER_RECV_BUF	; get address
	movf	SER_RECV_START,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	INDF0,A1_REG		; get byte
	incf	SER_RECV_START,W	; increment start pointer
	movwf	B1_REG
	movlw	SER_RECV_BUF_SIZE	; if B1_REG >= SER_RECV_BUF_SIZE, ...
	cpfslt	B1_REG
	clrf	B1_REG			; ..., B1_REG = 0
	movff	B1_REG,SER_RECV_START	; modify start pointer atomically
; put received byte into message send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x22			; put data into send buffer
	movwf	POSTINC0
	movff	A1_REG,POSTINC0		; data byte
; try again (for next byte)
	bra	SER_TASK



;===========================================================================================
; global variables
;===========================================================================================



;===========================================================================================
; global procedures
;===========================================================================================

	global		SER_INIT, SER_INT_RECV, SER_INT_TRANS, SER_TASK
	global		SER_SET_BAUD, SER_SEND



	end

